/*
  Una classe del tutto autonoma che rappresenta una generica macchina connessa a internet.
  (non solo chi partecipa a gaia) 
  Non fornisce nessun metodo all' esterno oltre a accetta()

  dentro un if lock di sicuro non verra' cambiato stato a false e non vengono chiamati metodi che
  possono deadlockkare
*/


package trz.gaia;

import java.util.*;

class Cellula extends Thread {

  static int N = 3;
  
  //come valori uso il numero massimo di connessioni
  public final static int EMBRIONE = 0; //cerca e accetta
  public static int FORTEMENTE_INSTABILE = N-2; //cerca e accetta
  public static int INSTABILE = N-1;  ///accetta ma non cerca
  public static int STABILE = N;  //non accetta e non cerca

  static float stabilita = 0.9991F;  //1 = non cade mai, 0 cade sempre
  static float rinascita = 0.9910F;  //1 = non rinasce mai, 0 rinasce subito
  static int pausa = 100;
  static boolean showCon = false;

  Internet net = null;
  Vector listaCon = new Vector();   //elenco delle connessioni aperte 
  Vector listaCell = new Vector();   //elenco delle cellule a cui e' connessa 

  Object lockAttiva = new Object();  //lock su attiva
  Object lockFermo = new Object();  //lock su fermo
  Object lockLocked = new Object();  //lock su fermo

  static Random random = new Random(); 

  protected String nome = null;
  int ip = -1;
  
  boolean attiva = false;   //partecipa a gaia
  boolean locked = false;   //monitor per l' aggiunta di connessioni

  boolean fermo = true; //per sospendere il thread

  CellulaView view = null;  //parte grafica della cellula
  
  static long tempoPing = 1000;
  long tempo = 0; //conteggio del tempo trascorso per il ping


  SearchNode nodo = null;
                         
  public Cellula(String nome, int ip, Internet net) {
  
    this(nome, ip, net, false);
  }

  public Cellula(String nome, int ip, Internet net, boolean attiva) {

    setPriority(Thread.MIN_PRIORITY); //in modo da lasciare piu' spazio alla gui 
                                      //(e' meglio mettere max la GUI ?)
       
    this.nome = nome;
    this.ip = ip;
    this.net = net;
    this.attiva = attiva;

    view = new CellulaView(this);
    
    nodo = new SearchNode(nome, nome);
    nodo.setAttivo(attiva);

    Gaia.graph.put(nodo);
  }

  public static void setN(int newN) {

    N = newN;
    FORTEMENTE_INSTABILE = N-2;
    INSTABILE = N-1;
    STABILE = N;
  }
  
  public boolean isAttiva() {
  
    synchronized(lockAttiva) {
      return attiva;
    }
  } 

  public void setAttiva(boolean attiva) {
    
    synchronized(lockAttiva) {
      this.attiva = attiva;
      nodo.setAttivo(attiva);
    }
    
    view.update();
    if (!isAttiva()) {
      //trace("mi resetto");
      reset();
    }
    //else
      //trace("riparto");
  } 

  //deve essere chiamato dentro un lock
  protected void reset() {

    Connessione con = null;
    Cellula tmp = null;
    
    synchronized (listaCon) {
      for (int i = 0; i < listaCon.size(); i++) {
        
        //trace("segno come chiusa connessione " + listaCon.elementAt(i));
        con = (Connessione)listaCon.elementAt(i);
        con.close();
        
        tmp = con.from;   //non so se io sono il from o il to di questa con 
        if (this == con.from)
          tmp = con.to;
        
        nodo.removeLink(tmp.nodo);
      }
      listaCon.setSize(0);
      listaCell.setSize(0);
    }
    view.update();
  }

  boolean isLocked() {
    synchronized(lockLocked) {
      return locked;
    }
  }

  //true se e' riuscito ad acquisire il lock
  boolean lock() {

    synchronized(lockLocked) {
      return locked? false: (locked = true);
    }
  } 

  void unlock() {

    synchronized(lockLocked) { 
      locked = false;
    }
  } 

  boolean isFermo() {
    synchronized(lockFermo) {
      return fermo;
    }
  }

  void setFermo(boolean fermo) {

    synchronized(lockFermo) {
      this.fermo = fermo;
    }
    view.update();
  } 

  //metodo che viene chiamato dalle cellule attive in cerca di nuove connessioni
  public boolean accetta(Connessione con) {
  
    //trace("mi viene offerta: " + con);

    if (lock()) { //prendo il lock

      synchronized(listaCon) {

        if (!isAttiva()) {
          unlock();
          return false;
        }

        if (listaCon.size() < STABILE) {
          //trace("accetto: " + con);
          addConnessione(con, con.from);
          unlock();
          return true;
        }
        unlock();
      }
    }
    return false;
  }

  public void run() {
  
    //scanna il range di ip e apre connessioni
    main:while(true) {

      while (isFermo())
        try {sleep(pausa);}catch(InterruptedException e){}

      while (!isAttiva()) {
        //controllare se ritorna attiva
        if (random.nextFloat() > rinascita) {
          synchronized(listaCon) {
            setAttiva(true);
          }
        }
        if (isFermo())
          continue main;
        try {sleep(pausa);}catch(InterruptedException e){}
      }

      while (listaCon.size() >= INSTABILE) { //non faccio nulla
        //controllare se cade
        if (random.nextFloat() > stabilita) {
          synchronized(listaCon) {
            setAttiva(false);
          }   //se non riesco a lockkare pace
          break;
        }

        checkConnections();

        if (isFermo())
          continue main;
        try {sleep(pausa);}catch(InterruptedException e){}
      }

      Cellula cellula = null;
      for (int ip = 0; ip < net.max && listaCon.size() < INSTABILE; ip++) {  //scorro tutto il range

        if (isFermo())  //male che vada il for gira una volta in piu'. 
          continue main;
    
        checkConnections();

        //controllare se diventa inattiva
        if (random.nextFloat() > stabilita) {
          synchronized(listaCon) {
            setAttiva(false);
          } //se non ruesco a mettere il lock "perdo il giro"
          continue main;
        }
      
        if (this.ip == ip)
          continue; 
          
        cellula = net.contatta(ip);

        //sarebbe meglio fare il controllo sull' ip evitando
        //di contattarla, ma adesso non c' ho voglia
        if (listaCell.contains(cellula))  //evito di connettermi alla stessa cellula
          continue;

        //trace("ho trovato cellula: " + cellula);
        proponi(cellula);

        if (showCon)
          net.view.disegnaConnessioni();

        try {   //faccio una pausa
          sleep(pausa*2);
        }catch(InterruptedException e){}
      
        //faccio in modo che il range di ip sia "infinito"
        if (ip == net.max-1)
          ip = -1;
      } //fine for
    }
  }  

  //questo metodo viene chiamato dentro un lock
  void proponi(Cellula cellula) {

    if (lock()) {
    
      synchronized (listaCon) {

        if (!isAttiva()) {
          unlock();
          return;
        }
    
        if(listaCon.size() < INSTABILE) {
          
          Connessione con = new Connessione(this, cellula);
          if (cellula.accetta(con)) {
            //trace(cellula + " ha accettato la connessione: " + con);
            addConnessione(con, cellula);
          }
          //else
          //  trace(cellula + " ha rifiutato la connessione: " + con);
        }
        //else
        //  trace("ho gia' abbastanza con");

      } //fine lock listaCon
      unlock();
    }   //fine if lock()
  }

  //questo metodo se ne sbatte di tutti
  void checkConnections() {

    if (System.currentTimeMillis() - tempo < tempoPing)
      return;
    else
      tempo = System.currentTimeMillis();  

    //trace("inizio check con");
    
    synchronized(listaCon) {

      Connessione con = null;
      Cellula tmp = null;
      for (int i = 0; listaCon.size() > 0 && i < listaCon.size();) {

        //trace(listaCon + " - " + listaCon.size() + " - " + i);

        con = (Connessione)listaCon.elementAt(i); 
        if (con.isClosed()) {
          
          //trace("chiudo la connessione " + con);
          listaCon.removeElementAt(i);

    
          tmp = con.from;   //non so se io sono il from o il to di questa con 
          if (this == con.from)
            tmp = con.to;
          listaCell.removeElementAt(listaCell.indexOf(tmp));

          nodo.removeLink(tmp.nodo);
        }
        else
          i++;
      }      
    }
    view.update();
    
    //trace("fine check con");
  }

  void addConnessione(Connessione con, Cellula cellula) {
 
    //trace("aggiungo la connessione " + con);
    synchronized(listaCon) {
      listaCon.addElement(con);
      listaCell.addElement(cellula);

      nodo.addLink(cellula.nodo);
    }    
    view.update();
  }
  
  //trovargli un lock
  public static void setStabilita(float stabilita) {
    Cellula.stabilita = stabilita;
  } 
  
  //trovargli un lock
  public static void setRinascita(float rinascita) {
    Cellula.rinascita = rinascita;
  } 
  
  //trovargli un lock
  public static void setVelocita(int pausa) {
    Cellula.pausa = pausa;
  } 

  //trovargli un lock
  public static void setTempoPing(int tempoPing) {
    Cellula.tempoPing = tempoPing;
  } 

  public String toString() {
  
    //return "[" + nome + "-" + listaCon.size() + "-" + (isAttiva()?"A":"I") + "-" + (isLocked()?"L":"U") + "]"; 
    return "["  + nome + "-" 
                + listaCon.size() + "-" 
                + (isAttiva()?"A":"I") + "-" 
                + (isFermo()?"F":"R") + "-" 
                + (isAlive()?"V":"M") + "]"; 
  }

  public String dump() {
  
    //String dump =  "[" + nome + "-" + listaCon.size() + "-" + (isAttiva()?"A":"I") + "-" + (isLocked()?"L":"U") + "]" + "\n";
    String dump =  "["  + nome + "-" 
                        + listaCon.size() + "-" 
                        + (isAttiva()?"A":"I") + "-" 
                        + (isFermo()?"F":"R") + "-" 
                        + (isAlive()?"V":"M") + "]" + "\n";
    
    for (int i = 0; i < listaCon.size(); i++)
      dump += "\t" + listaCon.elementAt(i) + "\n";
    return dump; 
  }
  
  public void trace(String s) {
  
    System.out.println(this + "> " + s);
  }
}
 
